
#ifndef __CLIENTSHELL_DE_H__
#define __CLIENTSHELL_DE_H__


	#include "basedefs_de.h"



	/////////////////////////////////////////////////////////////////////
	// ClientShell interface. This is full of callback functions that
	// DirectEngine will call to notify you of things.
	/////////////////////////////////////////////////////////////////////

	typedef struct ClientShellDE_t* LPCLIENTSHELLDE;

	// ---------------------------------------------------------------------- //
	// InitSoundInfo
	//
	// Used to initialize the sound engine.  Structure filled out by client shell.  
	// ---------------------------------------------------------------------- //

	#define INITSOUNDINFO_MAXSTATICVOICES	128
	#define INITSOUNDINFO_MAXVOICETYPES		128

	typedef struct InitSoundInfo_t
	{
		// Specifies number of static and dynamic voices.  Static voice slots allocate voice slots
		// for sound types.  Dynamic voice slots are given on first come basis.
		unsigned short				m_nNumStaticVoices;
		unsigned short				m_nNumDynamicVoices;

		// Array specifying voice types for the static voice slots.  Array must have m_nNumStaticVoices
		// number of slots.
		unsigned char 				m_StaticVoiceTypes[INITSOUNDINFO_MAXSTATICVOICES];

		// Number of voice types used.
		// Voice types values will be 0 through m_nNumVoiceTypes - 1, so it would be good to keep
		// this number low...
		unsigned char				m_nNumVoiceTypes;

		// Priority among voice types.  Allows a voice type with high priority to override lower
		// one.  This is used in overriding sounds in the dynamic slots.  Must have 
		// m_nNumVoiceTypes entries.
		unsigned char				m_VoiceTypePriorityTable[INITSOUNDINFO_MAXVOICETYPES];

		// Specifies, in percent, how much a control can change before the sound is remixed.  
		// Value of 0 means always remix.
		unsigned short				m_nTolerance;

		// Scaling factor should be in units of meters per game unit.  This value is used
		// for doppler shift...
		float						m_fDistanceFactor; 

		// Scales the affect of rolloff.
		float						m_fRolloffFactor; 

		// Scales the affect of doppler shift.  Value of zero turns doppler shift off.
		float						m_fDopplerFactor; 

		// Output sound format.  Number of channels (1 or 2), Sample rate (8000, 11025, 22050 or
		// 44100 kHz), Bits per sample (8 or 16)...
		unsigned short				m_nChannels;
		unsigned short				m_nSampleRate;
		unsigned short				m_nBitsPerSample;


	} InitSoundInfo;


	typedef struct ClientShellDE_t
	{

		// This is so DirectEngine will skip over a C++ object's VTable.
		#ifdef COMPILE_WITH_C 
			void		*cpp_4BytesForVTable;
		#endif 

		
		// Called after the engine has fully initialized and is ready to go!
		void		(*OnEngineInitialized)(LPCLIENTSHELLDE pShell);

		// Called before the engine uninitializes itself.
		void		(*OnEngineTerm)(LPCLIENTSHELLDE pShell);

		// Notification for when commands go on and off.
		void		(*OnCommandOn)(LPCLIENTSHELLDE pShell, int command);
		void		(*OnCommandOff)(LPCLIENTSHELLDE pShell, int command);

		// Key up/down notification. Try to use command notification whenever
		// possible because key up/down notification isn't portable.
		void		(*OnKeyDown)(LPCLIENTSHELLDE pShell, int key, int rep);
		void		(*OnKeyUp)(LPCLIENTSHELLDE pShell, int key);

		// Message handler.
		void		(*OnMessage)(LPCLIENTSHELLDE pShell, DBYTE messageID, HMESSAGEREAD hMessage);

		// Called when a model hits a model key.
		void		(*OnModelKey)(LPCLIENTSHELLDE pShell, HLOCALOBJ hObj, ArgList *pArgList);

		// Called right before a world is loaded.  After this call, it shows whatever
		// you've drawn to the screen, then loads the world.
		void		(*PreLoadWorld)(LPCLIENTSHELLDE pShell, char *pWorldName);

		// Called when you enter/exit a world.  You lose ALL server objects when
		// you exit the world, so don't reference them after exiting the world!
		void		(*OnEnterWorld)(LPCLIENTSHELLDE pShell);
		void		(*OnExitWorld)(LPCLIENTSHELLDE pShell);

		// Called when the server tells the client about a special effect object.
		void		(*SpecialEffectNotify)(LPCLIENTSHELLDE pShell, HLOCALOBJ hObj, HMESSAGEREAD hMessage);

		// Called when an object with client flag CF_NOTIFYONREMOVE is removed.
		void		(*OnObjectRemove)(LPCLIENTSHELLDE pShell, HLOCALOBJ hObj);

		void		(*PreUpdate)(LPCLIENTSHELLDE pShell);
		void		(*Update)(LPCLIENTSHELLDE pShell);
		void		(*PostUpdate)(LPCLIENTSHELLDE pShell);


		// Asks the client how the sounds should be setup...
		// The default sound setup is:
		// m_nNumStaticVoices			= 0
		// m_nNumDynamicVoices			= 32
		// m_pStaticVoiceTypes			= DNULL
		// m_nNumVoiceTypes				= 0
		// m_pVoiceTypePriorityTable	= DNULL
		// m_nTolerance					= 5
		// m_fDistanceFactor			= 1.0f
		// m_fRolloffFactor				= 1.0f
		// m_fDopplerFactor				= 1.0f
		// m_nChannels					= 2
		// m_nSampleRate				= 22050
		// m_nBitsPerSample				= 16
		void		(*SetupSound)( LPCLIENTSHELLDE pShell, InitSoundInfo *pSoundInitInfo );

		// This is set after you create and return your ClientShell.
		struct ClientDE_t *m_pClientDE;
	
	} ClientShellDE;


	#include "client_de.h"


	#define CLIENTSHELL_VERSION	1

	typedef ClientShellDE* (*CreateClientShellFn)();
	typedef void (*DeleteClientShellFn)(ClientShellDE *);
	
	// You must implement these two functions to create and delete your client shell.
	// When you create it, you should set all of its function pointers.
	ClientShellDE* CreateClientShell();
	void DeleteClientShell(ClientShellDE *);


	#define SETUP_CLIENTSHELL()\
		BEGIN_EXTERNC() \
			__declspec(dllexport) void GetClientShellFunctions(CreateClientShellFn *pCreate, DeleteClientShellFn *pDelete);\
			__declspec(dllexport) int GetClientShellVersion();\
		END_EXTERNC()\
		void GetClientShellFunctions(CreateClientShellFn *pCreate, DeleteClientShellFn *pDelete)\
		{\
			*pCreate = CreateClientShell;\
			*pDelete = DeleteClientShell;\
		}\
		int GetClientShellVersion()\
		{\
			return CLIENTSHELL_VERSION;\
		}


#endif  // __CLIENTSHELL_DE_H__

