
// server_de defines the ServerDE structure, which consists of all the
// server-side DirectEngine functionality.

#ifndef __SERVER_DE_H__
#define __SERVER_DE_H__

	
	#include "basedefs_de.h"
	#include "serverobj_de.h"


	DEFINE_HANDLE_TYPE(HCONVAR);
	DEFINE_HANDLE_TYPE(HCLIENTREF);



	// ---------------------------------------------------------------------- //
	// Client info flags.
	// ---------------------------------------------------------------------- //

	#define CIF_LOCAL		1	// Client is local (non-network) connection.
	#define CIF_PLAYBACK	2	// Virtual client from a demo playback.
								

	// ---------------------------------------------------------------------- //
	// Object states.
	// ---------------------------------------------------------------------- //
								
	#define OBJSTATE_ACTIVE			0	// Normal healthy object.
	#define OBJSTATE_INACTIVE		1	// Inactive (no updates, physics, or touch notifies).
	#define OBJSTATE_INACTIVE_TOUCH	2	// Inactive, but gets touch notifies.



	// ---------------------------------------------------------------------- //
	// Server state flags.
	// ---------------------------------------------------------------------- //
	
	#define SS_PAUSED		(1<<0)	// Server is paused.
	#define SS_DEMOPLAYBACK	(1<<1)	// We're running in 'demo playback' mode (read only).
	#define SS_LASTFLAG		SS_DEMOPLAYBACK

	
	// ---------------------------------------------------------------------- //
	// Object lists.
	// ---------------------------------------------------------------------- //
	
	typedef struct ObjectLink_t
	{
		HOBJECT		m_hObject;
		struct ObjectLink_t	*m_pNext;
	} ObjectLink;
	
	typedef struct ObjectList_t
	{
		ObjectLink	*m_pFirstLink;
		int			m_nInList;
	} ObjectList;


	// ---------------------------------------------------------------------- //
	// Standing info.
	// ---------------------------------------------------------------------- //

	typedef struct StandingInfo_t
	{
		
		// The plane of whatever you're standing on.
		DPlane		m_Plane;

		// The object you're standing on.
		HOBJECT		m_hObject;
	
	} StandingInfo;



	// ---------------------------------------------------------------------- //
	// GenericProp -
	// Here's a list of which property types map to which variables:
	// PT_STRING   - m_Vec and m_Color if it's formatted like '1 2 3', m_String, m_Long, m_Float, m_Bool.
	//             - if this is 'true' or 'false', it'll set m_Long, m_Float, and m_Bool to 0 or 1.

	// PT_VECTOR   - m_Vec, m_Color, m_String ('1 2 3')
	// PT_COLOR    - m_Vec, m_Color, m_String ('1 2 3')
	// PT_REAL     - m_String, m_Long, m_Float, m_Bool
	// PT_FLAGS    - m_String, m_Long, m_Float, m_Bool
	// PT_BOOL     - m_String, m_Long, m_Float, m_Bool
	// PT_LONGINT  - m_String, m_Long, m_Float, m_Bool
	// PT_ROTATION - m_Rotation and m_Vec (with euler rotation).
	// ---------------------------------------------------------------------- //

	#define MAX_GP_STRING_LEN 200

	typedef struct GenericProp_t
	{
		union
		{
			DVector	m_Vec;
			DVector	m_Color;
		};

		char		m_String[MAX_GP_STRING_LEN+1];
		DRotation	m_Rotation;
		long		m_Long;
		float		m_Float;
		DBOOL		m_Bool;
	} GenericProp;


	// ---------------------------------------------------------------------- //
	// ---------------------------------------------------------------------- //
	// Sound effects.
	// 
	// Structure filled out by code playing sound.  Sound is destroyed in
	// the following ways:
	//
	//		not looping, no handle
	//			not attached
	//				Client destroys when sound finished.
	//			attached
	//				Client destroys when object destroyed.
	//
	//		handle returned
	//			Client does not destroy, must be destroyed by creator.
	//
	// The type of sounds are:  ambient, local and 3D.  The flags controlling
	// these types are mutually exclusive.  An ambient sound will have distance
	// effects of rolloff, but no orientation effects.  A local sound will have
	// no orientation or distance effects and will be as if the sound was played
	// inside the player's head.  The 3d sound will have distance, orientation
	// and doppler effects.
	//
	// Sounds are played from the beginning once the message reaches the client.
	// If it is important that the playback be synchronized with the server's 
	// beginning time, then set PLAYSOUND_TIMESYNC.  Normally, this is not
	// that important.  The server will use it internally if a client comes in
	// after a sound has already been played.  It does this on looping sounds.
	//
	// ---------------------------------------------------------------------- //

	#define PLAYSOUND_LOCAL			0x0000	// Play sound locally (inside head)
	#define PLAYSOUND_AMBIENT		0x0001	// Play sound as ambient sound.
	#define PLAYSOUND_3D			0x0002	// Play sound as 3D sound.
	#define PLAYSOUND_LOOP			0x0004	// Loop the sound.
	#define PLAYSOUND_ATTACHED		0x0008	// Attached to object m_hObject
	#define PLAYSOUND_GETHANDLE		0x0010	// Handle requested
	#define PLAYSOUND_TIME			0x0020	// Server must time sound
	#define PLAYSOUND_CTRL_VOL		0x0040	// Control volume m_nVolume
	#define PLAYSOUND_CTRL_FREQ		0x0080	// Control frequency m_wFrequency
	#define PLAYSOUND_CTRL_PAN		0x0100	// Control pan m_wPan
	#define PLAYSOUND_CTRL_ORIENT	0x0200	// Control orientation.  Must have PLAYSOUND_3D and PLAYSOUND_CTRL_VOL also.
	#define PLAYSOUND_CLIENT		0x0400	// Client side sound
	#define PLAYSOUND_TIMESYNC		0x0800	// Playback synchronized with server clock
	#define PLAYSOUND_FILESTREAM	0x1000	// Stream the file

	typedef struct PlaySoundInfo_t
	{
		// PLAYSOUND flags.
		DDWORD		m_dwFlags;
	
		// File name of sound
		char 		m_szSoundName[_MAX_PATH];

		// Object sound is attached to.
		// Only needed if PLAYSOUND_ATTACHED set
		HOBJECT		m_hObject;

		// Handle of sound
		// Filled by PlaySound if PLAYSOUND_GETHANDLE set
		HSOUNDDE	m_hSound;

		// Voice type
		// Value must be 0 through InitSoundIno::m_nNumVoiceTypes - 1
		unsigned char m_nVoiceType;

		// Voice priority within voice type m_nVoiceType
		// 0 is lowest priority
		unsigned char m_nPriority;

		// Maximum radius of sound.  No sound outside this.
		// Only needed if PLAYSOUND_3D or PLAYSOUND_AMBIENT
		float		m_fOuterRadius;
		
		// Inner radius of sound.  Sound at maximum volume inside this radius
		// Only needed if PLAYSOUND_3D or PLAYSOUND_AMBIENT
		float		m_fInnerRadius;
		
		// Volume of sound in percent [0,100].
		// Only needed if PLAYSOUND_CTRL_VOL set, otherwise defaults 100
		// If PLAYSOUND_3D set, then this is the inside cone volume
		DBYTE		m_nVolume;

		// Frequency adjustment multiplier (0.0,MAX_FLOAT)
		// Only needed if PLAYSOUND_LOCAL and PLAYSOUND_CTRL_FREQ set
		// If PLAYSOUND_CTRL_FREQ clear, then defaults 1.0
		float		m_fFrequencyScale;

		// Pan in percent [-100(left),100(right)]
		// Only needed if PLAYSOUND_LOCAL and PLAYSOUND_CTRL_FREQ set
		// If PLAYSOUND_CTRL_PAN clear, then defaults 0
		signed char	m_iPan;

		// Position of sound.
		// Only needed if PLAYSOUND_AMBIENT and PLAYSOUND_3D set
		// Only needed if PLAYSOUND_ATTACHED cleared
		DVector		m_vPosition;

		// Forward orientation of sound.
		// Only needed if PLAYSOUND_3D set
		// Only needed if PLAYSOUND_ATTACHED cleared.  If attached, uses object's orientation.
		// Only needed if PLAYSOUND_CTRL_ORIENT set
		DVector		m_vOrientation;

		// Sound cone arc length in degrees [0,360]
		// Only needed if PLAYSOUND_3D set
		// Must have PLAYSOUND_CTRL_VOL set, otherwise defaults 360
		// Only needed if PLAYSOUND_CTRL_ORIENT set
		D_WORD		m_wConeAngle;

		// Outside of cone volume scale in percent [0,100]
		// Only needed if PLAYSOUND_3D set
		// Must have PLAYSOUND_CTRL_VOL set, otherwise defaults 100
		// Only needed if PLAYSOUND_CTRL_ORIENT set
		DBYTE		m_nOutsideConeVolume;

	} PlaySoundInfo;

	#define PLAYSOUNDINFO_COPY( dest, src ) \
		memcpy(( void * )&( dest ), ( void * )&( src ), sizeof( PlaySoundInfo ));

	#define PLAYSOUNDINFO_INIT(x) memset(&x, 0, sizeof(x));\


	// ---------------------------------------------------------------------- //
	// Intersection stuff.
	// ---------------------------------------------------------------------- //

	// Return DTRUE to select this object and DFALSE to not select it.
	typedef DBOOL (*ObjectFilterFn)(HOBJECT hObj, void *pUserData);


	// Pass this in to the IntersectSegment routine.
	typedef struct IntersectQuery_t
	{
		// Start and end points.
		DVector			m_From;
		DVector			m_To;

		// Only used for CastRay.. this is the direction the ray should be cast in.
		// This doesn't need to be normalized.
		DVector			m_Direction;

		// A combination of the intersect flags (in de_codes.h).
		DDWORD			m_Flags;
		
		// If this is not NULL, then it'll call this function when it has a 
		// POSSIBLE object intersection (it doesn't know if it really intersects
		// yet when it calls this).  If you return FALSE from this function, 
		// then it will ignore the object and continue on.
		ObjectFilterFn	m_FilterFn;
		
		// Passed into pUserData of the filter function.
		void			*m_pUserData;		
	} IntersectQuery;

	
	typedef struct IntersectInfo_t
	{
		// Point of intersection.
		DVector		m_Point;

		// Plane of intersection.
		DPlane		m_Plane;

		// Object it hit.
		HOBJECT		m_hObject;

		// The polygon it hit (if it's a world poly).
		// Value is INVALID_HPOLY if it didn't hit one.
		HPOLY		m_hPoly;

		// Surface flags of what it hit (these aren't object flags, and are
		// only set if it hit the world or a WorldModel).
		DDWORD		m_SurfaceFlags;
	} IntersectInfo;



	// ---------------------------------------------------------------------- //
	// CServerDE interface. This is your interface to most DirectEngine
	// functionality. 
	// ---------------------------------------------------------------------- //

	typedef struct ServerDE_t
	{
		
		// Main functions.

			// Returns DNULL if the class doesn't exist.
			HCLASS		(*GetClass)(char *pName);

			// Get a class's static object.  Returns DE_NOTFOUND if the server
			// doesn't have a static object of this class.
			DRESULT		(*GetStaticObject)(HCLASS hClass, HOBJECT *obj);

			// Returns the world's object..
			HOBJECT		(*GetWorldObject)();

			// Get an object's class.
			HCLASS		(*GetObjectClass)(HOBJECT hObject);

			// Tells if hClass is the class of hTest (or is derived from hTest).
			// If you pass in DNULL for hTest (like if GetClass returns DNULL in 
			// IsKindOf(hClass, GetClass("TestClass"))), it will return FALSE.
			DBOOL		(*IsKindOf)(HCLASS hClass, HCLASS hTest);
			

			// Creates an object using the object's default object type.
			LPBASECLASS	(*CreateObject)(HCLASS hClass, struct ObjectCreateStruct_t *pStruct);

			// Creates an object using string to specify properties.
			LPBASECLASS	(*CreateObjectProps)(HCLASS hClass, struct ObjectCreateStruct_t *pStruct, char *pszProps );

			// Remove the object from the world.  Note: the object won't be removed
			// or deleted until the end of the frame.
			void		(*RemoveObject)(HOBJECT hObject);


			// Returns the current time, in seconds, since the server started up.
			DFLOAT		(*GetTime)();
			
			// Returns the time since the last server shell update.
			DFLOAT		(*GetFrameTime)();


			// Access the global force vector applied to all objects (mainly for gravity).
			DRESULT		(*GetGlobalForce)(DVector *pVec);
			DRESULT		(*SetGlobalForce)(DVector *pVec);

			
			// Server state flags.  (Combination of the SS_ flags above).
			// SetServerFlags returns the new flags (some may not be accepted).
			DDWORD		(*GetServerFlags)();
			DDWORD		(*SetServerFlags)(DDWORD flags);
			
			
		// Helpers..
			
			// Returns a random number in the given range.
			DFLOAT		(*Random)(DFLOAT min, DFLOAT max);

			// Returns an integer number in the given range.
			int			(*IntRandom)(int min, int max);

			// Returns a number from 0 to scale.
			DFLOAT		(*RandomScale)(DFLOAT scale);

			// Only use this for debugging.. sends an STC_BPRINT message
			// with the string in it.
			void		(*BPrint)(char *pMsg, ...);

			// A better BPrint.. prints right to the console (doesn't do any network stuff) and 
			// doesn't matter if you're connected yet.
			void		(*CPrint)(char *pMsg, ...);

			// Get/Set the sky definition.
			DRESULT		(*GetSkyDef)(struct SkyDef_t *pDef);
			DRESULT		(*SetSkyDef)(struct SkyDef_t *pDef);

			// Add/Remove objects from the sky list.
			// Each object should have a unique index.
			DRESULT		(*AddObjectToSky)(HOBJECT hObj, DDWORD index);
			DRESULT		(*RemoveObjectFromSky)(HOBJECT hObj);


		// String functions.  Strings are reference counted objects that cannot
		// be manipulated.  When you create one with FormatString or CreateString,
		// the reference count is 1.  When you copy a string with CopyString, the
		// reference count is incremented.  When you free one with FreeString,
		// it decrements the reference count.. when the reference count goes to
		// zero, it deletes the string.  If you forget to free up any strings, 
		// DirectEngine will spit out a message telling you about it..

			// In Windows, messageCode comes from your resource DLL.  The messages
			// need to be formatted with %1!s! %2!s! (the number is the argument 
			// number and the !s! says its a string).  You can also use !f! and !d!
			// for floating point and whole number.
			HSTRING		(*FormatString)(int messageCode, ...);
			
			// Copy a string.. much more efficient than CreateString().
			HSTRING		(*CopyString)(HSTRING hString);
			HSTRING		(*CreateString)(char *pString);
			void		(*FreeString)(HSTRING hString);

			DBOOL		(*CompareStrings)(HSTRING hString1, HSTRING hString2);
			DBOOL		(*CompareStringsUpper)(HSTRING hString1, HSTRING hString2);

			// Get the string's data.. you really should only use this for strings
			// that you stored off and want to pass to the engine for a filename 
			// or something..  Most strings could be in some format other than ANSI.
			char*		(*GetStringData)(HSTRING hString);


		// World control.

			// Portal flags are a combination of PORTAL_ flags in de_codes.
			// Returns DE_OK, DE_NOTFOUND, or DE_NOTINWORLD.  This is
			// case sensitive.
			DRESULT		(*GetPortalFlags)(char *pPortalName, DDWORD *pFlags);
			DRESULT		(*SetPortalFlags)(char *pPortalName, DDWORD flags);

			// Intersect a line segment.. (used to be raycast, but line segments are WAY faster).
			// Returns TRUE and fills in pInfo if it hit something.
			DBOOL		(*IntersectSegment)(IntersectQuery *pQuery, IntersectInfo *pInfo);

			// Same as IntersectSegment, except for it casts a ray from pQuery->m_From
			// in the direction of pQuery->m_Dir.
			DBOOL		(*CastRay)(IntersectQuery *pQuery, IntersectInfo *pInfo);
			
			///////////////// NOT IMPLEMENTED YET /////////////////
			// Find out what's at a given point (is it inside the world, outside, what
			// area brushes is it inside..)  You must give back the list to the engine
			// with RelinquishList()!
			ObjectList*	(*GetPointAreas)(DVector *pPoint);

			///////////////// NOT IMPLEMENTED YET /////////////////
			// Find the objects intersecting this box.
			ObjectList*	(*GetBoxIntersecters)(DVector *pMin, DVector *pMax);

			// Get the shade (RGB, 0-255) at the point you specify.
			// Returns DFALSE if the point is outside the world.
			DBOOL		(*GetPointShade)(DVector *pPoint, DVector *pColor);
			
			// Gets all the objects with the given name.
			// Don't forget to relinquish the list you get!
			ObjectList*	(*FindNamedObjects)(char *pName);

			// Find all the objects touching the given sphere.
			// Don't forget to relinquish the list you get!
			// Be VERY careful of using a large radius.  This function can take
			// tons of cycles if the radius is too large.
			ObjectList* (*FindObjectsTouchingSphere)(DVector *pPosition, float radius);

			// Finds objects touching the cone you specify.  This will provide a slight
			// overestimation of the cone you specify.  This can be much faster than
			// FindObjectsTouchingSphere, depending on how small the angle is.
			ObjectList* (*FindObjectsTouchingCone)(DVector *pPos, DVector *pDir, float angle, float dist);
			
			// Any time the engine gives you a list, you must 'give it back' with this,
			// or else you'll have tons of memory leaks.
			void		(*RelinquishList)(ObjectList *pList);


		// Object list management.

			// Allocate a list (free it with RelinquishList).
			ObjectList*	(*CreateObjectList)();
			
			// Add objects to a list.
			ObjectLink*	(*AddObjectToList)(ObjectList *pList, HOBJECT hObj);
			
			// Remove an object from the list given the handle. 
			// Note: DirectEngine will just do a search over all the objects in the
			//       list for this, so be warned.
			void		(*RemoveObjectFromList)(ObjectList *pList, HOBJECT hObj);



		// Messaging.

			// Use this to setup a special effect message.  If your object has
			// a special effect message, the client shell's SpecialEffectNotify() 
			// will be called.  An object can only have ONE special effect message.
			// Returns DNULL if the object is invalid.
			// If the object already has a special effect message, then it
			// clears out the current one.
			// Special effect messages can be up to 64 bytes in length.
			HMESSAGEWRITE	(*StartSpecialEffectMessage)(LPBASECLASS pObject);

			// Starts an instant special effect message.  The message is sent to 
			// any clients that can see the position you specify.
			HMESSAGEWRITE	(*StartInstantSpecialEffectMessage)(DVector *pPos);

			// Use this to start a message to an object, then use the WriteToMessage
			// functions and EndMessage to finish it off.
			HMESSAGEWRITE	(*StartMessageToObject)(LPBASECLASS pSender, HOBJECT hSendTo, DDWORD messageID);

			// Send a message to the server shell.  pSender can be NULL.
			DRESULT			(*StartMessageToServer)(LPBASECLASS pSender, DDWORD messageID, HMESSAGEWRITE *hWrite);

			// Start a message, put data into it (no more than 50 bytes), and
			// end the message..  You can only have one HMESSAGEWRITE open at a time.
			// If you set hSendTo to NULL, it'll send the message to all clients.
			HMESSAGEWRITE	(*StartMessage)(HCLIENT hSendTo, DBYTE messageID);
			
			DRESULT			(*WriteToMessageFloat)(HMESSAGEWRITE hMessage, float val);
			DRESULT			(*WriteToMessageByte)(HMESSAGEWRITE hMessage, DBYTE val);
			DRESULT			(*WriteToMessageDWord)(HMESSAGEWRITE hMessage, DDWORD val);
			DRESULT			(*WriteToMessageString)(HMESSAGEWRITE hMessage, char *pStr);
			DRESULT			(*WriteToMessageVector)(HMESSAGEWRITE hMessage, DVector *pVal); // 12 bytes
			DRESULT			(*WriteToMessageCompVector)(HMESSAGEWRITE hMessage, DVector *pVal); // 9 bytes
			DRESULT			(*WriteToMessageRotation)(HMESSAGEWRITE hMessage, DRotation *pVal);
			DRESULT			(*WriteToMessageHString)(HMESSAGEWRITE hMessage, HSTRING hString);
			
			// Helper so you don't have to FormatString and FreeString..
			DRESULT			(*WriteToMessageFormattedHString)(HMESSAGEWRITE hMessage, int messageCode, ...);

			// Note: you can't send object references to the client yet, so the client can't
			// even read object references.
			DRESULT			(*WriteToMessageObject)(HMESSAGEWRITE hMessage, HOBJECT hObj);	
			
			DRESULT			(*EndMessage)(HMESSAGEWRITE hMessage);

			// When your OnMessage function is called, use the handle you're given
			// to read the message data with these functions.
			float			(*ReadFromMessageFloat)(HMESSAGEREAD hMessage);
			DBYTE			(*ReadFromMessageByte)(HMESSAGEREAD hMessage);
			DDWORD			(*ReadFromMessageDWord)(HMESSAGEREAD hMessage);
			char*			(*ReadFromMessageString)(HMESSAGEREAD hMessage);
			void			(*ReadFromMessageVector)(HMESSAGEREAD hMessage, DVector *pVal); // 12 bytes
			void			(*ReadFromMessageCompVector)(HMESSAGEREAD hMessage, DVector *pVal); // 9 bytes
			void			(*ReadFromMessageRotation)(HMESSAGEREAD hMessage, DRotation *pVal);
			HOBJECT			(*ReadFromMessageObject)(HMESSAGEREAD hMessage);
			HSTRING			(*ReadFromMessageHString)(HMESSAGEREAD hMessage);
			
			// Reset reading (so you can read the message again).
			// This is useful if you read out of a message and subclasses
			// will be reading out of it.  Note: the message will AUTOMATICALLY
			// reset when you hit the end, so you won't need this in most cases.
			void			(*ResetRead)(HMESSAGEREAD hRead);

		
		// Client Functions.

			// Attach one client to another.  This means that messages sent to hParent
			// will also go to hClient, and hClient's client object will be reported
			// as hParent's object.  This is useful for demo playback when you want
			// to watch another player's view.
			DRESULT	(*AttachClient)(HCLIENT hParent, HCLIENT hChild);

			// Detach a client from its parent attachment.
			DRESULT	(*DetachClient)(HCLIENT hClient);

			// Iterate over the clients.  Pass in NULL to start.  Returns NULL
			// when there are no more.
			HCLIENT	(*GetNextClient)(HCLIENT hPrev);

			// Iterate over the saved client references.  These come from when
			// a world is saved or from objects with FLAG_KEEPALIVE set when
			// you switch worlds.  Pass in DNULL to start.  See the Client Functions
			// section for the functions to look at these.
			HCLIENTREF	(*GetNextClientRef)(HCLIENTREF hRef);

			// HCLIENTREFs are used for comparison to HCLIENTs.
			DDWORD	(*GetClientRefInfoFlags)(HCLIENTREF hClient);
			DBOOL	(*GetClientRefName)(HCLIENTREF hClient, char *pName, int maxLen);
			HOBJECT	(*GetClientRefObject)(HCLIENTREF hClient);
			
			// Get a client's (unique) ID.
			DDWORD	(*GetClientID)(HCLIENT hClient);
			
			// Get a client's name.
			DBOOL	(*GetClientName)(HCLIENT hClient, char *pName, int maxLen);
			
			// A combination of the CIF_ flags above.
			DDWORD	(*GetClientInfoFlags)(HCLIENT hClient);

			// User data for HCLIENTs.
			void	(*SetClientUserData)(HCLIENT hClient, void *pData);
			void*	(*GetClientUserData)(HCLIENT hClient);

			// Kick a client off the server.
			// OnRemoveClient will be called.
			void	(*KickClient)(HCLIENT hClient);


		// The 'game console state'.  This is a set of console variables and functions
		// that is used internally by the application DLLs.  The user can't access this
		// at all.  It's very useful for tracking global game variables and triggering
		// them through game objects.

			// Run a string through the game console.  The syntax is the same as any
			// other console string (semicolons separate commands, and new variable names
			// create new HCONVARs).
			void	(*RunGameConString)(char *pString);

			// Sets a console variable.  Creates a new one if it doesn't exist.
			void	(*SetGameConVar)(char *pName, char *pVal);
			
			// Get a handle to a console variable.
			HCONVAR	(*GetGameConVar)(char *pName);
			
			// Get the floating point value of a console variable.  If the hVar is
			// NULL, then it returns 0.
			float	(*GetVarValueFloat)(HCONVAR hVar);

			// Get the string value of a console variable.  If the hVar is
			// NULL, then it returns NULL.
			char*	(*GetVarValueString)(HCONVAR hVar);


		// Helpers.

			// These are used to get the property values from the world file.
			// The property names are case sensitive.  If the property doesn't exist,
			// it will return DE_NOTFOUND.
			DRESULT		(*GetPropString)(char *pPropName, char *pRet, int maxLen);
			DRESULT		(*GetPropVector)(char *pPropName, DVector *pRet);
			DRESULT		(*GetPropColor)(char *pPropName, DVector *pRet);
			DRESULT		(*GetPropReal)(char *pPropName, float *pRet);
			DRESULT		(*GetPropFlags)(char *pPropName, DDWORD *pRet);
			DRESULT		(*GetPropBool)(char *pPropName, DBOOL *pRet);
			DRESULT		(*GetPropLongInt)(char *pPropName, long *pRet);
			DRESULT		(*GetPropRotation)(char *pPropName, DRotation *pRet);
			DRESULT		(*GetPropRotationEuler)(char *pPropName, DVector *pAngles); // pAngles = (pitch, yaw, roll)
			
			// Fills in the GenericProp for the different data types.  For a list
			// of which property types map to which GenericProp variable, see 
			// the GenericProp structure.
			// Note: if the property exists, it always initializes the members in the prop first,
			// so if the GenericProp variable doesn't support that property type, 
			// it'll be zero (or ROT_INIT'd).
			DRESULT		(*GetPropGeneric)(char *pPropName, GenericProp *pProp);


			// See if a property exists.  If so, pPropType is filled in (if it's non-NULL) with
			// on of the PT_ defines.  Returns DE_NOTFOUND if it doesn't exist.
			DRESULT		(*DoesPropExist)(char *pPropName, int *pPropType);

			// Get handles from objects and objects from handles. Note: HandleToObject will return
			// NULL if the given object doesn't reside on this server (it's controlled remotely).
			HOBJECT		(*ObjectToHandle)(LPBASECLASS pObject);
			LPBASECLASS	(*HandleToObject)(HOBJECT hObject);

			// Tests if a command is on for this client.
			DBOOL	(*IsCommandOn)(HCLIENT hClient, int command);

			// Gets orientation vectors, given an angle vector (x, y, and z rotations).
			void	(*GetRotationVectors)(DRotation *pRotation, 
				DVector *pUp, DVector *pRight, DVector *pForward);

			// Does a (fast) upper-case string comparison of the 2 strings.
			DBOOL	(*UpperStrcmp)(char *pStr1, char *pStr2);

			// Rotate this rotation around the given axis by a certain amount.
			void	(*RotateAroundAxis)(DRotation *pRotation, DVector *pAxis, float amount);

			// Treat the rotation like Euler angles...
			void	(*EulerRotateX)(DRotation *pRotation, float amount);
			void	(*EulerRotateY)(DRotation *pRotation, float amount);
			void	(*EulerRotateZ)(DRotation *pRotation, float amount);

			// Align a rotation to a normal and spin it.
			// Use pUp to set the frame of reference.
			// If pUp is NULL or the same direction as pVector, then DirectEngine 
			// will kludge an up vector.
			void	(*AlignRotation)(DRotation *pRotation, DVector *pVector, DVector *pUp);

			// Setup the rotation based on euler angles 
			// pitch = rotation around X
			// yaw = rotation around Y
			// roll = rotation around Z
			DRESULT	(*SetupEuler)(DRotation *pRotation, float pitch, float yaw, float roll);

			// Interpolate between two rotations (with quaternions).
			DRESULT	(*InterpolateRotation)(DRotation *pDest, DRotation *pRot1, DRotation *pRot2, float t);

			// Find out what the object is standing on.
			void	(*GetStandingOn)(HOBJECT hObj, StandingInfo *pInfo);

			
			// Create a rotation matrix for the given rotation.
			void	(*SetupRotationMatrix)(DMatrix *pMat, DRotation *pRot);
			
			// Create a translation matrix for the given translation.
			void	(*SetupTranslationMatrix)(DMatrix *pMat, DVector *pTranslation);

			// Rotate around a point.
			void	(*SetupRotationAroundPoint)(DMatrix *pMat, DRotation *pRot, DVector *pPoint);

			// Parse a string for arguments.  Works similar to command line parser.
			// args within quotation marks are considered one arg.
			// Returns the all args after pCommand and before end of string or semi-colon.
			//		pCommand - Beginning of string.
			//		pNewCommandPos - Parse fills this in with the ending position
			//		argBuffer - memory Parse can use as scratchpad
			//		argPointers - Parse fills these pointers in to point to args found.
			//		nArgs - Parse fills in the number of arguments found.
			//		Return - DTRUE, semicolon found, more args in string.  DFALSE, no more args.
			int			(*Parse)(char *pCommand, char **pNewCommandPos, char *argBuffer, char **argPointers, int *nArgs);

		// Sound functions.

			// Play a sound with full control
			// Arguments:
			//		pPlaySoundInfo - sound control structure
			// Returns:
			//		DE_OK if successful.
			//		DE_ERROR on error.
			DRESULT	(*PlaySound)( PlaySoundInfo *pPlaySoundInfo );

			// Play sound without effects (inside players head), sound removed when finished playing.
			// Arguments:
			//		pSoundName - relative path to sound name including ext.
			//		nType - Type of voice, must be 0 through InitSoundIno::m_nNumVoiceTypes - 1
			//		nPriority - priority of sound compared to others of same type
			// Returns:
			//		DE_OK if successful.
			//		DE_ERROR on error.
			DRESULT	(*PlaySoundLocal)(char *pSoundName, unsigned char nType, unsigned char nPriority );

			// Play sound with full 3D effects
			// Arguments:
			//		hObject - Handle to object sound is attached to.  Sound removed when object removed.
			//		pSoundName - relative path to sound name including ext.
			//		fOuterRadius - Max radius of sound
			//		fInnerRadius - Inner radius of sound.
			//		bLooping - 1: Loop the sound, 0: no loop.  If not looping, sound removed when sound finishes playing.
			//		nType - Type of voice, must be 0 through InitSoundIno::m_nNumVoiceTypes - 1
			//		nPriority - priority of sound compared to others of same type
			// Returns:
			//		DE_OK if successful.
			//		DE_ERROR on error.
			DRESULT	(*PlaySound3D)(HOBJECT hObject, char *pSoundName, float fOuterRadius, float fInnerRadius, DBYTE bLooping, unsigned char nType, unsigned char nPriority );

			// Play sound as looping ambient.  Has rolloff with distance, but no orientation
			// Arguments:
			//		hObject - Handle to object sound is attached to.  Sound removed when object removed.
			//		pSoundName - relative path to sound name including ext.
			//		fOuterRadius - Max radius of sound
			//		fInnerRadius - Inner radius of sound
			//		nType - Type of voice, must be 0 through InitSoundIno::m_nNumVoiceTypes - 1
			//		nPriority - priority of sound compared to others of same type
			// Returns:
			//		DE_OK if successful.
			//		DE_ERROR on error.
			DRESULT	(*PlaySoundAmbient)(HOBJECT hObject, char *pSoundName, float fOuterRadius, float fInnerRadius, unsigned char nType, unsigned char nPriority );

			// Get sound time.
			// Arguments:
			//		hSoundHandle - Handle to sound.
			// Returns:
			//		DE_OK if successful.
			//		DE_ERROR on error.
			float	(*GetSoundTimeRemaining)( HSOUNDDE hSound );

			// Kill a sound.
			// Arguments:
			//		hSoundHandle - Handle to sound.
			// Returns:
			//		DE_OK if successful.
			//		DE_ERROR on error.
			DRESULT	(*KillSound)( HSOUNDDE hSound );


		// Inter-object link manipulation.

			// If you want to hold on to an HOBJECT and get notification when the
			// object is removed from the world, you can create an inter-object link
			// between the two.  The owner will get MID_LINKBROKEN notification when 
			// the link is being broken (by either the owner or linked object being removed).
			void	(*CreateInterObjectLink)(HOBJECT hOwner, HOBJECT hLinked);

			// Breaks an inter-object link between the owner and linked object
			// (if one exists..)  You can only break a link from the owner, since
			// the linked object doesn't even know it's linked to the owner..
			// Note: MID_LINKBROKEN will NOT be called.
			void	(*BreakInterObjectLink)(HOBJECT hOwner, HOBJECT hLinked);


		// Object manipulation.

			// Attaches hChild to hParent.  If pNodeName is set, it'll attach hChild to a 
			// specific node on hParent.
			// DE will automatically detach if you remove hParent,
			// but it will NOT automatically detach if you remove hChild.  If you remove the 
			// child without removing the attachment, the results are undefined.
			// Returns DE_OK, DE_NODENOTFOUND, or DE_INVALIDPARAMS.
			DRESULT	(*CreateAttachment)(HOBJECT hParent, HOBJECT hChild, char *pNodeName, 
				DVector *pOffset, DRotation *pRotationOffset, HATTACHMENT *pAttachment);

			// Removes an attachment.  Note: an attachment becomes invalid when you remove the parent
			// so it'll crash if you call it with an attachment with a parent you've removed.
			DRESULT	(*RemoveAttachment)(HATTACHMENT hAttachment);

			// Get/Set object color info (RGBA 0-1).
			// All objects default to (0,0,0,1)
			// For lights, this changes the light's color.
			// For models, this brightens a model's shading.
			DBOOL	(*GetObjectColor)(HOBJECT hObject, float *r, float *g, float *b, float *a);
			DBOOL	(*SetObjectColor)(HOBJECT hObject, float r, float g, float b, float a);

			// Get/Set an object's user-defined flags.
			DDWORD	(*GetObjectUserFlags)(HOBJECT hObj);
			DRESULT	(*SetObjectUserFlags)(HOBJECT hObj, DDWORD flags);

			// Get/Set an object's mass (default is 30).
			float	(*GetObjectMass)(HOBJECT hObj);
			void	(*SetObjectMass)(HOBJECT hObj, float mass);

			// Get/Set force ignore limit.
			float	(*GetForceIgnoreLimit)(HOBJECT hObj);
			void	(*SetForceIgnoreLimit)(HOBJECT hObj, float limit);
			
			// Use this to iterate thru all the objects in the world.
			// Pass in NULL to start off with.  It'll return NULL when
			// you have iterated through all the objects.
			// This is generally a bad idea, but sometimes you have to.
			HOBJECT	(*GetNextObject)(HOBJECT hObj);

			// Same as GetNextObject, but this iterates over all the inactive objects.
			HOBJECT	(*GetNextInactiveObject)(HOBJECT hObj);
			
			// Tells you what type of object this is.
			short	(*GetObjectType)(HOBJECT hObj);

			// Get the object's name.
			char*	(*GetObjectName)(HOBJECT hObject);

			// Set an object's friction coefficient.
			void	(*SetFrictionCoefficient)(HOBJECT hObj, float coefficient);

			// This is a counter that controls when each object gets its Update()
			// function called. This is set to 0.0001 when an object is created 
			// so by default, Update() gets called right away.
			void	(*SetNextUpdate)(HOBJECT hObj, DFLOAT nextUpdate);

			// Sets the time which the engine will automatically deactivate an object 
			void	(*SetDeactivationTime)(HOBJECT hObj, DFLOAT fDeactivationTime);

			// Activates any objects seen by this object
			void	(*PingObjects)(HOBJECT hObj);

			// Object data accessors.
			void	(*GetObjectPos)(HOBJECT hObj, DVector *pos);
			void	(*SetObjectPos)(HOBJECT hObj, DVector *pos);  // Exactly the same as calling Teleport().

			// Scale the object (only works on models and sprites).
			void	(*ScaleObject)(HOBJECT hObj, DVector *pNewScale);
			DRESULT	(*GetObjectScale)(HOBJECT hObj, DVector *pScale);

			// This function moves an object, colliding/pushing/crushing objects
			// in its way (if it's solid..)
			void	(*MoveObject)(HOBJECT hObj, DVector *pNewPos);

			// Teleports the object directly to the position.
			void	(*TeleportObject)(HOBJECT hObj, DVector *pNewPos);

			// Get/Set the object's rotation.
			DRESULT	(*GetObjectRotation)(HOBJECT hObj, DRotation *pRotation);
			DRESULT	(*SetObjectRotation)(HOBJECT hObj, DRotation *pRotation);

			// Tilt their acceleration to be along the plane they're standing on.
			void	(*TiltToPlane)(HOBJECT hObj, DVector *pNormal);

			// Get/Set the object's flags.
			DDWORD	(*GetObjectFlags)(HOBJECT hObj);
			void	(*SetObjectFlags)(HOBJECT hObj, DDWORD flags);

			// Get/Set the object's state.  State is one of the OBJSTATE_ defines above.
			void	(*SetObjectState)(HOBJECT hObj, int state);
			int		(*GetObjectState)(HOBJECT hObj);

			// Get the object's current dimensions.
			void	(*GetObjectDims)(HOBJECT hObj, DVector *pNewDims);
			
			// Changes the object's dimensions without pushing against objects and world.
			// Requires you to know that the object won't collide with anything.  This is
			// faster than SetObjectDims2.
			// Returns 
			//		DE_OK if successful.
			//		DE_INVALIDPARAMS if invalid object such as OT_CONTAINER or OT_WORLDMODEL.
			DRESULT	(*SetObjectDims)(HOBJECT hObj, DVector *pNewDims);

			// Changes the object's dimensions.
			// Based on what it might have collided with, information is returned.  This isn't
			// as fast as SetObjectDims.
			// Returns 
			//		DE_OK if successful.
			//		DE_ERROR if could not grow to full dims. pNewDims is modified to show 
			//		actual dims.
			//		DE_INVALIDPARAMS if invalid object such as OT_CONTAINER or OT_WORLDMODEL.
			DRESULT	(*SetObjectDims2)(HOBJECT hObj, DVector *pNewDims);

			// Get/Set acceleration and velocity
			void	(*GetVelocity)(HOBJECT hObj, DVector *pVel);
			void	(*SetVelocity)(HOBJECT hObj, DVector *pVel);

			void	(*GetAcceleration)(HOBJECT hObj, DVector *pAccel);
			void	(*SetAcceleration)(HOBJECT hObj, DVector *pAccel);

			// Get/Set blocking priority (defaults to 0).
			// See FAQ for a description of how this works.
			void	(*SetBlockingPriority)(HOBJECT hObj, DBYTE pri);
			DBYTE	(*GetBlockingPriority)(HOBJECT hObj);


		// Sprite manipulation.

			// This clips the sprite on the poly.
			// Returns DE_OK or DE_ERROR if not a sprite.
			// Pass in INVALID_HPOLY to un-clip the sprite.
			DRESULT	(*ClipSprite)(HOBJECT hObj, HPOLY hPoly);

		
		// Light manipulation.

			// Get/Set a light's color (RGB, 0.0f to 1.0f).
			// When you create a light, its color defaults to (0,0,0).
			
			// Note: a light's color is snapped to 256 different values, so if
			// you want to do any fancy interpolation or anything, you'll need
			// to store your own, higher precision, color values.
			
			// Note: this just calls GetObjectColor/SetObjectColor.
			void	(*GetLightColor)(HOBJECT hObj, float *r, float *g, float *b);
			void	(*SetLightColor)(HOBJECT hObj, float r, float g, float b);

			// Get/Set a light's radius.
			// When you create a light, its radius defaults to 100.
			float	(*GetLightRadius)(HOBJECT hObj);
			void	(*SetLightRadius)(HOBJECT hObj, float radius);

		
		// Model manipulation.

			// Iterate through the model's nodes.  Returns DE_FINISHED when there are no more.
			// hCurNode = INVALID_MODEL_NODE;
			// while(interface->GetNextModelNode(hModel, hCurNode, &hCurNode) == DE_OK)
			// { ... }
			DRESULT	(*GetNextModelNode)(HOBJECT hObject, HMODELNODE hNode, HMODELNODE *pNext);

			// Get a model node's name.
			DRESULT	(*GetModelNodeName)(HOBJECT hObject, HMODELNODE hNode, char *pName, DDWORD maxLen);

			// Get a model's command string.
			DRESULT	(*GetModelCommandString)(HOBJECT hObj, char *pStr, DDWORD maxLen);

			// Hide/unhide a node on the model (they're all unhidden by default).
			// Returns DE_OK, DE_ERROR, LT_NOCHANGE, or DE_NODENOTFOUND.
			DRESULT	(*SetModelNodeHideStatus)(HOBJECT hObj, char *pNodeName, DBOOL bHidden);
			DRESULT	(*GetModelNodeHideStatus)(HOBJECT hObj, char *pNodeName, /* out */DBOOL *bHidden);
			
			// Get the current (global) transformation for a model node.
			// Returns DFALSE if the node does not exist or if the object
			// you pass in is not a model.
			DBOOL	(*GetModelNodeTransform)(HOBJECT hObj, char *pNodeName,	
				DVector *pPos, DRotation *pRot);

			// Get an animation index from a model.
			// Returns -1 if the animation doesn't exist (or if the object isn't a model).
			HMODELANIM	(*GetAnimIndex)(HOBJECT hObj, char *pAnimName);

			// If the object is a model, this sets its current animation.
			void	(*SetModelAnimation)(HOBJECT hObj, HMODELANIM hAnim);

			// Returns the animation the model is currently on.  (HMODELANIM)-1 if none.
			HMODELANIM	(*GetModelAnimation)(HOBJECT hObj);

			// Starts the current animation over.
			DRESULT	(*ResetModelAnimation)(HOBJECT hObj);

			// Tells what the playback state of the model is (a combination of the
			// MS_ bits defined in basedefs_de.h).
			DDWORD	(*GetModelPlaybackState)(HOBJECT hObj);

			// Get/Set the looping state of the model.  The default state is TRUE.
			void	(*SetModelLooping)(HOBJECT hObj, DBOOL bLoop);
			DBOOL	(*GetModelLooping)(HOBJECT hObj);

			// Get the model filenames.  You can pass in DNULL for pFilename and pSkinName.
			// Returns DFALSE if the object is not a model.
			// Initializes pFilename or pSkinName to zero length if the model doesn't have 
			// a filename or skin.
			DBOOL	(*GetModelFilenames)(HOBJECT hObj, char *pFilename, int fileBufLen, char *pSkinName, int skinBufLen);

			// Change the model's filenames.
			void	(*SetModelFilenames)(HOBJECT hObj, char *pFilename, char *pSkinName);

			// Get the endpoints of the model's extents for its current animation frame.
			// This is in world space so if you just want the dimensions, take (pMax - pMin) / 2.
			// Returns DE_ERROR if it's not a model.
			DRESULT	(*GetModelFrameBox)(HOBJECT hObj, DVector *pMin, DVector *pMax);

			// Gets a model animation's user dimensions (what they set for dimensions in ModelEdit).
			DRESULT	(*GetModelAnimUserDims)(HOBJECT hObj, DVector *pDims, HMODELANIM hAnim);


		// Container manipulation.
		
			// Get the container's container code (can only be set in the ObjectCreateStruct during creation).
			// Returns DFALSE if the object isn't a container.
			DBOOL	(*GetContainerCode)(HOBJECT hObj, D_WORD *pCode);

			// Gets the list of containers the object is inside.
			// pFlagList must be the same size as pContainerList.
			// Returns the number of elements in pList filled in.
			DDWORD	(*GetObjectContainers)(HOBJECT hObj, 
				HOBJECT *pContainerList, DDWORD *pFlagList, DDWORD maxListSize);

			// Gets the list of objects inside the container.
			// Returns the number of elements in pList filled in.
			DDWORD	(*GetContainedObjects)(HOBJECT hContainer, 
				HOBJECT *pObjectList, DDWORD *pFlagList, DDWORD maxListSize);

		
		// Surface manipulation.

			// Get the texture flags from a poly.  Returns DE_OK
			// or DE_ERROR if no world is loaded or hPoly is invalid.
			DRESULT	(*GetPolyTextureFlags)(HPOLY hPoly, DDWORD *pFlags);

		// Save game

			DRESULT (*SaveObjects)( char *pszSaveFileName, ObjectList *pList, DDWORD dwParam );

			DRESULT (*RestoreObjects)( char *pszRestoreFileName, DDWORD dwParam );

		// Load world

			DRESULT (*LoadWorld)( char *pszWorldFileName, DBOOL bLoadWorldObjects, DBOOL bRunWorld );

			DRESULT (*RunWorld)();

	} ServerDE;



#endif  // __SERVER_DE_H__



