
// --------------------------------
// File    : ClientPlayer.cpp
// Function: Client-side class for controlling a player object and a camera.
//         : Fairly simplistic.
// --------------------------------


#include "de_client.h"
#include "clientplayer.h"

// --------------------------------------------------------------- //
// Command IDs (mapped to keys in autoexec.cfg).
// --------------------------------------------------------------- //

#define COMMAND_LEFT			10
#define COMMAND_RIGHT			11
#define COMMAND_LOOKUP			15
#define COMMAND_LOOKDOWN		16
#define COMMAND_QUIT			17
#define COMMAND_CAMERA_LEFT		20
#define COMMAND_CAMERA_RIGHT	21
#define COMMAND_CAMERA_TOGGLE	22
#define COMMAND_CAMERA_ZOOMIN	23
#define COMMAND_CAMERA_ZOOMOUT	24
#define COMMAND_CAMERA_RAISE	25
#define COMMAND_CAMERA_LOWER	26

ClientPlayer::ClientPlayer()
{
	m_hName = DNULL;
	m_hCamera = DNULL;
	m_pClient = DNULL;
}

ClientPlayer::~ClientPlayer()
{
	// Free our HSTRING for the name.
	if (m_pClient)
	{
		if (m_hName)
			m_pClient->FreeString(m_hName);
		m_hName = DNULL;
	}
}

// ----------------
// Init
//		Initializes the player with a pointer to the client.
// ----------------
void ClientPlayer::Init(ClientDE *pClient)
{
	HCONSOLEVAR hVar;

	m_pClient = pClient;

	// Read our playername in.  We're not really using this anywhere, it's just
	// an example of reading in custom strings via console variables.
	hVar = m_pClient->GetConsoleVar("playername");

	if (hVar) {
		m_hName = m_pClient->CreateString(m_pClient->GetVarValueString(hVar));
	} else {
		m_hName = m_pClient->CreateString("Player");
	}

	m_Pitch = 0.0f;
	m_Yaw = 0.0f;
	m_Roll = 0.0f;

	m_LokiCam.Init(m_pClient);
}

// ----------------
// Update
//		Main update loop.
// ----------------
void ClientPlayer::Update()
{
	HLOCALOBJ hClientObj;
	float radiansPerSec = 2.5f;
    DRotation rRot;
	HMESSAGEWRITE hMessage;

	// Get our client object
	hClientObj = m_pClient->GetClientObject();

	// If we're in the world, let's update!
	if(hClientObj)
	{
		// Update our orientation.  We do this on the client and send the
		// angles to the server so mouse orientation is zero latency.
		if(m_pClient->IsCommandOn(COMMAND_LEFT))
		{
			m_Yaw -= radiansPerSec * m_pClient->GetFrameTime();
		}
		if(m_pClient->IsCommandOn(COMMAND_RIGHT))
		{
			m_Yaw += radiansPerSec * m_pClient->GetFrameTime();
		}

		// Handle our various Camera commands
		if (m_pClient->IsCommandOn(COMMAND_CAMERA_ZOOMIN))
		{
			m_LokiCam.Zoom(-5.0f);
		}
		if (m_pClient->IsCommandOn(COMMAND_CAMERA_ZOOMOUT))
		{
			m_LokiCam.Zoom(5.0f);
		}
		if (m_pClient->IsCommandOn(COMMAND_CAMERA_RAISE))
		{
			m_LokiCam.UpdateOffset(5.0f);
		}
		if (m_pClient->IsCommandOn(COMMAND_CAMERA_LOWER))
		{
			m_LokiCam.UpdateOffset(-5.0f);
		}
		if (m_pClient->IsCommandOn(COMMAND_CAMERA_LEFT))
		{
			m_LokiCam.UpdateYaw(radiansPerSec * m_pClient->GetFrameTime() * -1.0f);
		}
		if (m_pClient->IsCommandOn(COMMAND_CAMERA_RIGHT))
		{
			m_LokiCam.UpdateYaw(radiansPerSec * m_pClient->GetFrameTime());
		}
		if (m_pClient->IsCommandOn(COMMAND_LOOKUP))
		{
			m_LokiCam.UpdatePitch(radiansPerSec * m_pClient->GetFrameTime() * -1.0f);
		}
		if (m_pClient->IsCommandOn(COMMAND_LOOKDOWN))
		{
			m_LokiCam.UpdatePitch(radiansPerSec * m_pClient->GetFrameTime());
		}

		// Update our player's rotation.
		m_pClient->SetupEuler(&rRot, 0.0f, m_Yaw, 0.0f);
		m_pClient->SetObjectRotation(hClientObj,&rRot);

		// Perform our camera update
		m_LokiCam.Update();

		// Send the player orientation to the server.  The server will update the
		// player object rotation.  Note that '1' is an arbitrary number defined
		// within this demo.  For an actual game, you would probably wish to define
		// somewhat more detailed message handlers.  See the message handler in the
		// object.dll source to see where this message is processed.
		hMessage = m_pClient->StartMessage(1);
		m_pClient->WriteToMessageRotation(hMessage, &rRot);
		m_pClient->EndMessage(hMessage);
	}
}

// ----------------
// ResetCamera
//		Resets the Camera
// ----------------
void ClientPlayer::ResetCamera()
{
	m_LokiCam.ResetCamera();
}

// ----------------
// OnCommandOn
//		Called from the clientshell OnCommandOn to handle any commands that
//		the clientshell didn't.  Really only handles the camera toggle.
// ----------------
void ClientPlayer::OnCommandOn(int command)
{
	if (command == COMMAND_CAMERA_TOGGLE)
	{
		m_LokiCam.CycleMode();
	}
}