
// --------------------------------
// File    : LokiCam.cpp
// Function: Cheap hack camera class, provided as an example of various types
//         : of camera modes.  All are intended more as starting points for
//         : user extensions.  Partly inspired by Ashley Matheson's ChaseCam
//         : tutorial.
//         :
//         : Don't blame Ash, though. :)
// --------------------------------


#include "de_client.h"
#include "lokicam.h"

#include "windows.h"
#include "winbase.h"

#define CAMERA_MOVEMENT_LEFT		0
#define CAMERA_MOVEMENT_RIGHT		1
#define CAMERA_MOVEMENT_UP			2
#define CAMERA_MOVEMENT_ZOOM		3

#define CAMERA_MOVEMENT_FINAL		4

// Constructor
LokiCam::LokiCam()
{
	VEC_SET(m_CameraPos,0.0f,0.0f,0.0f);
	m_CameraZoom = 100.0f;
	m_CameraMode = LOKICAM_MODE_RELATIVE;
	m_hTarget = DNULL;
	m_hCamera = DNULL;	
}

// Destructor
LokiCam::~LokiCam()
{
	// Blah
}

// ----------------
// Init
//		Initializes the LokiCam with a pointer to the client.
// ----------------
void LokiCam::Init(ClientDE *pClient)
{
	m_pClientDE = pClient;
	VEC_SET(m_CameraPos,0.0f,0.0f,0.0f);

	// Sanity check
	if (m_pClientDE)
		m_pClientDE->SetupEuler(&m_CameraRotation,0.0f,0.0f,0.0f);

	m_CameraZoom = 100.0f;
	m_CameraMode = LOKICAM_MODE_RELATIVE;
	m_hTarget = DNULL;
	m_hCamera = DNULL;

	m_hTarget = pClient->GetClientObject();
}

// ----------------
// ResetCamera
//		Resets the Camera.
// ----------------
void LokiCam::ResetCamera()
{
	VEC_SET(m_CameraPos,0.0f,0.0f,0.0f);

	// Sanity check
	if (m_pClientDE)
		m_pClientDE->SetupEuler(&m_CameraRotation,0.0f,0.0f,0.0f);

	m_CameraZoom = 100.0f;
	m_CameraMode = LOKICAM_MODE_RELATIVE;
}

// ----------------
// CycleMode
//		Cycles to the next camera mode.
// ----------------
void LokiCam::CycleMode()
{
	if (m_CameraMode++ >= LOKICAM_MODE_LASTMODE)
		m_CameraMode = 0;

	// Tell us what mode we're in.
	switch(m_CameraMode)
	{
		case LOKICAM_MODE_RELATIVE:
			m_pClientDE->CPrint("Camera mode: Relative-locked");
			break;
		case LOKICAM_MODE_ISO:
			m_pClientDE->CPrint("Camera mode: Iso");
			break;
		case LOKICAM_MODE_GAUNTLET:
			m_pClientDE->CPrint("Camera mode: 'Gauntlet'");
			break;
		case LOKICAM_MODE_GAUNTLET2:
			m_pClientDE->CPrint("Camera mode: 'Gauntlet 2'");
			break;
		case LOKICAM_MODE_FREEFORM:
			m_pClientDE->CPrint("Camera mode: Freeform");
			break;
		default:
			m_pClientDE->CPrint("Camera mode: %d", m_CameraMode);
	}
}

// ----------------
// Update
//		Primary update loop.
// ----------------
void LokiCam::Update()
{
	// Sanity check that we have a camera.
	if (m_hCamera)
	{
		// Sanity check that we have a target.
		if (!m_hTarget)
		{
			m_pClientDE->CPrint("Camera: No target set, defaulting to player.");
			m_hTarget = m_pClientDE->GetClientObject();
		}

		// Find appropriate sub-update to call.
		switch(m_CameraMode)
		{
			case LOKICAM_MODE_RELATIVE:
				UpdateRelative();
				break;
			case LOKICAM_MODE_ISO:
				UpdateIso();
				break;
			case LOKICAM_MODE_GAUNTLET:
				UpdateGauntlet();
				break;
			case LOKICAM_MODE_GAUNTLET2:
				UpdateGauntlet2();
				break;
			case LOKICAM_MODE_FREEFORM:
				UpdateFreeform();
				break;
			default:
				m_pClientDE->CPrint("Camera: unknown camera mode!  Defaulting to relative.");
				m_CameraMode = LOKICAM_MODE_RELATIVE;
				m_CameraZoom = 100.0f;
				UpdateRelative();
		}

	} else 
	{
		m_pClientDE->CPrint("Camera: *bzzzt* sorry, no camera set, thank you for playing.");
	}
}

// ----------------
// UpdateRelative
//		Updates the camera in a position behind the player, locked relative to them.
//		This mode was inspired by Ashley Matheson's tutorial.
// ----------------
void LokiCam::UpdateRelative()
{
	DVector targetPos, cameraPos;
	DRotation targetRotation;
	DVector vUp, vRight, vForward;

	// Get some basic information about our target (probably the player)
	m_pClientDE->GetObjectPos(m_hTarget,&targetPos);
	m_pClientDE->GetObjectRotation(m_hTarget,&targetRotation);

	// Get the target's vectors
	m_pClientDE->GetRotationVectors(&targetRotation, &vUp, &vRight, &vForward); 

	// Since vForward is a unit vector, you can multiply by distance 
	// you want to move back..
	VEC_MULSCALAR(vForward, vForward, m_CameraZoom);  
	VEC_SUB(cameraPos, targetPos, vForward);
	cameraPos.y += 80;

	m_pClientDE->SetObjectRotation(m_hCamera,&targetRotation);
	m_pClientDE->SetObjectPos(m_hCamera,&cameraPos);

	m_CameraPos = cameraPos;
	m_CameraRotation = targetRotation;
}

// ----------------
// UpdateIso
//		This locks the camera in a vaguely 'Iso' mode (think 'Diablo').  It was written in
//		about five minutes as a demonstration to someone.
// ----------------
void LokiCam::UpdateIso()
{
	DVector targetPos, cameraPos;
	DRotation cameraRotation;
	DVector vUp, vRight, vForward;

	// Get some basic information about our target (probably the player)
	m_pClientDE->GetObjectPos(m_hTarget,&targetPos);

	// Set up our rotation
	m_pClientDE->SetupEuler(&cameraRotation,MATH_DEGREES_TO_RADIANS(45.0f),MATH_DEGREES_TO_RADIANS(45.0f),0.0f);

	// Get the camera's vectors
	m_pClientDE->GetRotationVectors(&cameraRotation, &vUp, &vRight, &vForward); 

	// Since vForward is a unit vector, you can multiply by distance 
	// you want to move back..
	VEC_MULSCALAR(vForward, vForward, m_CameraZoom);  
	VEC_SUB(cameraPos, targetPos, vForward);

	m_pClientDE->SetObjectRotation(m_hCamera,&cameraRotation);
	m_pClientDE->SetObjectPos(m_hCamera,&cameraPos);

	m_CameraPos = cameraPos;
	m_CameraRotation = cameraRotation;
}

// ----------------
// UpdateGauntlet
//		Locks the camera directly overhead of the player, looking downward.
//		Inspired, unsurprisingly, by Gauntlet.
// ----------------
void LokiCam::UpdateGauntlet()
{
	DVector targetPos, cameraPos;
	DRotation cameraRotation;
	DVector vUp, vRight, vForward;

	// Get some basic information about our target (probably the player)
	m_pClientDE->GetObjectPos(m_hTarget,&targetPos);

	// Set up our rotation
	m_pClientDE->SetupEuler(&cameraRotation,MATH_DEGREES_TO_RADIANS(90.0f),0.0f,0.0f);

	// Get the camera's vectors
	m_pClientDE->GetRotationVectors(&cameraRotation, &vUp, &vRight, &vForward); 

	// Since vForward is a unit vector, you can multiply by distance 
	// you want to move back..
	VEC_MULSCALAR(vForward, vForward, m_CameraZoom);  
	VEC_SUB(cameraPos, targetPos, vForward);

	m_pClientDE->SetObjectRotation(m_hCamera,&cameraRotation);
	m_pClientDE->SetObjectPos(m_hCamera,&cameraPos);

	m_CameraPos = cameraPos;
	m_CameraRotation = cameraRotation;
}

// ----------------
// UpdateGauntlet2
//		UpdateGauntlet, only slightly offset.
// ----------------
void LokiCam::UpdateGauntlet2()
{
	DVector targetPos, cameraPos;
	DRotation cameraRotation;
	DVector vUp, vRight, vForward;

	// Get some basic information about our target (probably the player)
	m_pClientDE->GetObjectPos(m_hTarget,&targetPos);

	// Set up our rotation
	m_pClientDE->SetupEuler(&cameraRotation,MATH_DEGREES_TO_RADIANS(75.0f),MATH_DEGREES_TO_RADIANS(15.0f),0.0f);

	// Get the camera's vectors
	m_pClientDE->GetRotationVectors(&cameraRotation, &vUp, &vRight, &vForward); 

	// Since vForward is a unit vector, you can multiply by distance 
	// you want to move back..
	VEC_MULSCALAR(vForward, vForward, m_CameraZoom);  
	VEC_SUB(cameraPos, targetPos, vForward);

	m_pClientDE->SetObjectRotation(m_hCamera,&cameraRotation);
	m_pClientDE->SetObjectPos(m_hCamera,&cameraPos);

	m_CameraPos = cameraPos;
	m_CameraRotation = cameraRotation;
}

// ----------------
// UpdateFreeform
//		A 'freeform' camera mode where the camera can be rotated around the
//		target freely.
//		Mostly for experimentation.
// ----------------
void LokiCam::UpdateFreeform()
{
	DVector targetPos, cameraPos;
	DRotation cameraRotation;
	DVector vUp, vRight, vForward;

	// Get some basic information about our target (probably the player)
	m_pClientDE->GetObjectPos(m_hTarget,&targetPos);

	// Set up our rotation
	m_pClientDE->SetupEuler(&cameraRotation,m_Pitch,m_Yaw,m_Roll);

	// Get the camera's vectors
	m_pClientDE->GetRotationVectors(&cameraRotation, &vUp, &vRight, &vForward); 

	// Since vForward is a unit vector, you can multiply by distance 
	// you want to move back..
	VEC_MULSCALAR(vForward, vForward, m_CameraZoom);  
	VEC_SUB(cameraPos, targetPos, vForward);
	cameraPos.y += m_Offset;

	m_pClientDE->SetObjectRotation(m_hCamera,&cameraRotation);
	m_pClientDE->SetObjectPos(m_hCamera,&cameraPos);

	m_CameraPos = cameraPos;
	m_CameraRotation = cameraRotation;
}

// ----------------
// Zoom
//		Changes the camera's zoom by the amount in delta.
// ----------------
void LokiCam::Zoom(float delta)
{
	m_CameraZoom += delta;

	if (m_CameraZoom <= 60.f) m_CameraZoom = 60.0f;
	if (m_CameraZoom >= 400.0f) m_CameraZoom = 400.0f;
}

// ----------------
// UpdateOffset
//		Updates the Offset (for freeform) by the amount in delta.
// ----------------
void LokiCam::UpdateOffset(float delta)
{
	m_Offset += delta;

	if (m_Offset <= 0.0f) m_Offset = 0.0f;
	if (m_Offset >= 200.0f) m_Offset = 200.0f;
}

// ----------------
// UpdatePitch
//		Updates the Pitch (for freeform) by the amount in delta.
// ----------------
void LokiCam::UpdatePitch(float delta)
{
	m_Pitch += delta;
}

// ----------------
// UpdateYaw
//		Updates the Yaw (for freeform) by the amount in delta.
// ----------------
void LokiCam::UpdateYaw(float delta)
{
	m_Yaw += delta;
}

// ----------------
// UpdateRoll
//		Updates the Roll (for freeform) by the amount in delta.
// ----------------
void LokiCam::UpdateRoll(float delta)
{
	m_Roll += delta;
}


