
// --------------------------------
// File    : c_lensflare.c
// Function: C source code for actual LensFlare effect, wrapped by LensFlare.cpp.
//         : Provided as an example of how to manipulate engine data under C as
//         : well. :)
//         : Special thanks to Mike Dussault for the original base code for the flare.
// --------------------------------

#include "client_de.h"
#include "c_lensflare.h"

#include <stdio.h>

// ------------------------------------------------------------------ //
// Defines.
// ------------------------------------------------------------------ //

// 0th ring is always the flare center.
#define NUM_FLARE_RINGS	4

#define FLARE_VISIBLE_ANGLE	(MATH_HALFPI*0.5f)


// ------------------------------------------------------------------ //
// Structures.
// ------------------------------------------------------------------ //

typedef struct
{
	HLOCALOBJ m_hObj;
	HLOCALOBJ m_hFlareCenter;
	HLOCALOBJ m_Rings[NUM_FLARE_RINGS];
	HLOCALOBJ m_Burst;
	DBOOL     m_bVisible;
	DLink     m_Link;
} Flare;


// ----------------
// lf_MakeFlareInvisible
//		Hides the lensflare.
// ----------------
static void lf_MakeFlareInvisible(ClientDE *pClientDE, Flare *pFlare)
{
	int i;
	HCONSOLEVAR hVar;

	// Leave the flare 'burst' sprite itself visible, but make all the rings invisible.
	for(i=0; i < NUM_FLARE_RINGS; i++)
		pClientDE->SetObjectFlags(pFlare->m_Rings[i], 0);

	// Now check if lenshideburst is true.
	hVar = pClientDE->GetConsoleVar("lenshideburst");
	if (pClientDE->GetVarValueFloat(hVar) != 0.0f)
		pClientDE->SetObjectFlags(pFlare->m_Burst,0);

}

// ----------------
// lf_ToggleFlare
//		Disables or enables the flare.
// ----------------
static void lf_ToggleFlare(ClientDE *pClientDE, Flare *pFlare, DBOOL bShow)
{
	DDWORD dwFlags;
	int i;

	// What flags do we need?
	if (bShow == DTRUE) {
		dwFlags = FLAG_VISIBLE | FLAG_SPRITE_NOZ;
	} else {
		dwFlags = 0;
	}

	// Set our visibility variable
	pFlare->m_bVisible = bShow;

	// Set the flags on the rings
	for(i=0; i < NUM_FLARE_RINGS; i++)
		pClientDE->SetObjectFlags(pFlare->m_Rings[i], dwFlags);

	// And on the burst
	pClientDE->SetObjectFlags(pFlare->m_Burst, dwFlags);
}

// ----------------
// lf_HideFlare
//		Exported equivalent to ToggleFlare DFALSE.
// ----------------
void lf_HideFlare(ClientDE *pClientDE, HLENSFLARE hFlare)
{
	lf_ToggleFlare(pClientDE, (Flare *)hFlare, DFALSE);
}

// ----------------
// lf_CreateFlare
//		Creates the Lensflare
// ----------------
HLENSFLARE lf_CreateFlare(ClientDE *pClientDE, HLOCALOBJ hObj)
{
	Flare *pFlare;
	ObjectCreateStruct createStruct;
	int i;

	// Create the flare.
	pFlare = (Flare*)malloc(sizeof(Flare));
	pFlare->m_Link.m_pData = pFlare;
	pFlare->m_hObj = hObj;
	pFlare->m_bVisible = DTRUE;

	// Create the rings.
	INIT_OBJECTCREATESTRUCT(createStruct);
	createStruct.m_ObjectType = OT_SPRITE;
	createStruct.m_Flags = FLAG_VISIBLE | FLAG_SPRITE_NOZ;
	pClientDE->GetObjectPos(hObj,&createStruct.m_Pos);

	sprintf(createStruct.m_Filename,"sprites\\flareburst.spr");
	pFlare->m_Burst = pClientDE->CreateObject(&createStruct);

	sprintf(createStruct.m_Filename,"sprites\\flarering.spr");
	pFlare->m_Rings[0] = pClientDE->CreateObject(&createStruct);

	for(i=1; i < NUM_FLARE_RINGS; i++)
	{
		sprintf(createStruct.m_Filename,"sprites\\flareouter%d.spr",i);
		pFlare->m_Rings[i] = pClientDE->CreateObject(&createStruct);
	}
	
	return (HLENSFLARE)pFlare;
}

// ----------------
// lf_DestroyFlare
//		Destroys the lensflare
// ----------------
void lf_DestroyFlare(ClientDE *pClientDE, HLENSFLARE hFlare)
{
	Flare *pFlare;
	int i;

	pFlare = (Flare*)hFlare;
	for(i=0; i < NUM_FLARE_RINGS; i++)
		pClientDE->DeleteObject(pFlare->m_Rings[i]);

	pClientDE->DeleteObject(pFlare->m_Burst);

	free(pFlare);
}


void lf_UpdateFlare(ClientDE *pClientDE, HLENSFLARE hFlare, HLOCALOBJ hViewer)
{
	Flare *pFlare;
	DRotation rotation;
	DVector up, right, forward;
	DVector flarePos, ringEnd, viewerPos;
	DVector temp, ringSpaceBetween, curPos;
	HCONSOLEVAR hVar;
	DBOOL bVisible;
	float varVal;
	float xPos, yPos, angle, angleT, cosVisibleAngle;
	int i;

	// Get our flare pointer...
	pFlare = (Flare*)hFlare;

	// Is the "lensflare" console variable set true?
	hVar = pClientDE->GetConsoleVar("lensflare");
	varVal = pClientDE->GetVarValueFloat(hVar);
	bVisible = (varVal != 0.0f);

	// If we're different visibility than we were last time, hide us.
	if (pFlare->m_bVisible != bVisible) lf_ToggleFlare(pClientDE,hFlare,bVisible);

	// Only calculate if we're visible.
	if (bVisible) 
	{
		// Get orientations and stuff.
		pClientDE->GetObjectPos(pFlare->m_hObj, &flarePos);
		pClientDE->GetObjectPos(hViewer, &viewerPos);
		pClientDE->GetObjectRotation(hViewer, &rotation);
		pClientDE->GetRotationVectors(&rotation, &up, &right, &forward);

		cosVisibleAngle = (float)cos(FLARE_VISIBLE_ANGLE * 0.5f);

		// Figure out if we can see it.
		VEC_SUB(temp, flarePos, viewerPos);
		VEC_NORM(temp);
		angle = VEC_DOT(temp, forward);
		
		if(angle < cosVisibleAngle)
		{
			// We can't see the flare from here, make it invisible and
			// return.
			lf_MakeFlareInvisible(pClientDE, hFlare);
			return;
		}

		angleT = 2.5f / (angle / cosVisibleAngle);

		// Figure out the line from the flare to the ring position.
		VEC_SUB(temp, flarePos, viewerPos);
		xPos = VEC_DOT(right, temp) * angleT;
		yPos = VEC_DOT(up, temp) * angleT;

		VEC_MULSCALAR(temp, right, -xPos);
		VEC_COPY(ringEnd, temp);

		VEC_MULSCALAR(temp, up, -yPos);
		VEC_ADD(ringEnd, ringEnd, temp);

		VEC_ADD(ringEnd, ringEnd, viewerPos);

		// Now just linearly interpolate from the flare position to the ring end.
		VEC_SUB(temp, ringEnd, flarePos);
		VEC_DIVSCALAR(ringSpaceBetween, temp, (float)(NUM_FLARE_RINGS));

		// Now just move the rings and burst into place!
		VEC_COPY(curPos, flarePos);
		for(i=0; i < NUM_FLARE_RINGS; i++)
		{
			pClientDE->SetObjectFlags(pFlare->m_Rings[i], FLAG_VISIBLE | FLAG_SPRITE_NOZ);
			pClientDE->SetObjectPos(pFlare->m_Rings[i], &curPos);
			if (i == 0)
			{
				pClientDE->SetObjectFlags(pFlare->m_Burst, FLAG_VISIBLE | FLAG_SPRITE_NOZ);
				pClientDE->SetObjectPos(pFlare->m_Burst, &curPos);
			}

			VEC_ADD(curPos, curPos, ringSpaceBetween);
		}
	}
}

// ----------------
// lf_GetFlareLink
//		Gets the DLink.  Can be used to make this C class work
//		in linked lists directly.
// ----------------
struct DLink_t* lf_GetFlareLink(HLENSFLARE hFlare)
{
	return &((Flare*)hFlare)->m_Link;
}


